import { useRef, useState, useEffect } from 'react';

const defaultReturn = {
    isValid: false,
    value: '',
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

var getMultipleSelectValue = (options) => options.filter(({ selected }) => selected).map(({ value }) => value);

var isRadioInput = (type) => type === 'radio';

var isCheckBoxInput = (type) => type === 'checkbox';

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        const fieldValue = fields[name];
        if (!fieldValue)
            throw new Error(`Expected ${name} field is missing`);
        return getRadioValue(fieldValue.options).value;
    }
    if (type === 'select-multiple')
        return getMultipleSelectValue(options);
    if (isCheckBoxInput(type))
        return checked ? value || checked : false;
    return value;
}

var isString = (value) => value && typeof value === 'string';

function getFieldsValues(fields, filedName) {
    return Object.values(fields).reduce((previous, data) => {
        const { ref, ref: { name }, } = data;
        const value = getFieldValue(fields, ref);
        if (isString(filedName)) {
            return name === filedName ? value : previous;
        }
        if (Array.isArray(filedName)) {
            if (filedName.includes(name)) {
                previous[name] = value;
            }
        }
        else {
            previous[name] = value;
        }
        return previous;
    }, {});
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url', 'textarea'];
const VALIDATION_MODE = {
    onBlur: 'onBlur',
    onchange: 'onChange',
    onSubmit: 'onSubmit',
};

var getValueAndMessage = (item) => ({
    value: typeof item === 'object' && item.value ? item.value : item,
    message: typeof item === 'object' && item.message ? item.message : '',
});

var isEmptyObject = (values) => values && values.constructor === Object && Object.keys(values).length === 0;

var validateField = async ({ ref, ref: { type, value, name, checked }, options, required, maxLength, minLength, min, max, pattern, validate, }, fields) => {
    const error = {};
    const isRadio = isRadioInput(type);
    const isCheckBox = isCheckBoxInput(type);
    const isSelectOrInput = !isCheckBox && !isRadio;
    if (required &&
        ((isCheckBox && !checked) ||
            (isSelectOrInput && value === '') ||
            (isRadio && !getRadioValue(fields[name].options).isValid) ||
            (!type && !value))) {
        error[name] = {
            type: 'required',
            message: isString(required) ? required : '',
            ref: isRadio ? (fields[name].options || [{ ref: '' }])[0].ref : ref,
        };
        return error;
    }
    if ((min || max) && !STRING_INPUTS.includes(type)) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        const { value: maxValue, message: maxMessage } = getValueAndMessage(max);
        const { value: minValue, message: minMessage } = getValueAndMessage(min);
        if (type === 'number') {
            exceedMax = maxValue && valueNumber > maxValue;
            exceedMin = minValue && valueNumber < minValue;
        }
        else if (DATE_INPUTS.includes(type)) {
            if (typeof maxValue === 'string')
                exceedMax = maxValue && new Date(value) > new Date(maxValue);
            if (typeof minValue === 'string')
                exceedMin = minValue && new Date(value) < new Date(minValue);
        }
        if (exceedMax || exceedMin) {
            error[name] = Object.assign({}, error[name], { type: exceedMax ? 'max' : 'min', message: exceedMax ? maxMessage : minMessage, ref });
            return error;
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const { value: maxLengthValue, message: maxLengthMessage, } = getValueAndMessage(maxLength);
        const { value: minLengthValue, message: minLengthMessage, } = getValueAndMessage(minLength);
        const exceedMax = maxLength && value.toString().length > maxLengthValue;
        const exceedMin = minLength && value.toString().length < minLengthValue;
        if (exceedMax || exceedMin) {
            error[name] = Object.assign({}, error[name], { type: exceedMax ? 'maxLength' : 'minLength', message: exceedMax ? maxLengthMessage : minLengthMessage, ref });
            return error;
        }
    }
    if (pattern) {
        const { value: patternValue, message: patternMessage } = getValueAndMessage(pattern);
        if (patternValue instanceof RegExp && !patternValue.test(value)) {
            error[name] = Object.assign({}, error[name], { type: 'pattern', message: patternMessage, ref });
            return error;
        }
    }
    if (validate) {
        const fieldValue = isRadio ? getRadioValue(options).value : value;
        const validateRef = isRadio && options ? options[0].ref : ref;
        if (typeof validate === 'function') {
            const result = await validate(fieldValue);
            if (typeof result === 'string' && result) {
                error[name] = Object.assign({}, error[name], { type: 'validate', message: result, ref: validateRef });
                return error;
            }
            else if (typeof result === 'boolean' && !result) {
                error[name] = Object.assign({}, error[name], { type: 'validate', message: '', ref: validateRef });
                return error;
            }
        }
        else if (typeof validate === 'object') {
            const validationResult = await new Promise((resolve) => {
                const values = Object.entries(validate);
                values.reduce(async (previous, [key, validate], index) => {
                    const lastChild = values.length - 1 === index;
                    if (typeof validate === 'function') {
                        const result = await validate(fieldValue);
                        if (typeof result !== 'boolean' || !result) {
                            const data = {
                                type: key,
                                message: isString(result) ? result : '',
                                ref: validateRef,
                            };
                            return lastChild ? resolve(data) : data;
                        }
                    }
                    return lastChild ? resolve(previous) : previous;
                }, {});
            });
            if (validationResult && !isEmptyObject(validationResult)) {
                error[name] = Object.assign({}, error[name], { ref: validateRef }, validationResult);
                return error;
            }
        }
    }
    return error;
};

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    if (!ref.removeEventListener)
        return;
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

function findRemovedFieldAndRemoveListener(fields, validateWithStateUpdate, { ref, mutationWatcher, options }, forceDelete = false) {
    if (!ref || !ref.type)
        return;
    const { name, type } = ref;
    if (isRadioInput(type) && options) {
        options.forEach(({ ref }, index) => {
            if (ref instanceof HTMLElement && !document.body.contains(ref) && options && options[index]) {
                removeAllEventListeners(options[index], validateWithStateUpdate);
                (options[index].mutationWatcher || { disconnect: () => { } }).disconnect();
                options.splice(index, 1);
            }
        });
        if (Array.isArray(options) && !options.length) {
            delete fields[name];
            return fields;
        }
    }
    else if ((ref instanceof HTMLElement && !document.body.contains(ref)) || forceDelete) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

function attachEventListeners({ field, validateAndStateUpdate, isRadio, }) {
    const { ref } = field;
    if (!ref.addEventListener)
        return;
    ref.addEventListener(isCheckBoxInput(ref.type) || isRadio ? 'change' : 'input', validateAndStateUpdate);
    ref.addEventListener('blur', validateAndStateUpdate);
}

function parseErrorSchema(error) {
    return error.inner.reduce((previous, current, index) => {
        previous[current.path] = error.errors[index];
        return previous;
    }, {});
}
async function validateWithSchema(ValidationSchema, data) {
    try {
        await ValidationSchema.validate(data, { abortEarly: false });
        return {};
    }
    catch (e) {
        return parseErrorSchema(e);
    }
}

function combineFieldValues(data) {
    const output = Object.entries(data).reduce((previous, [key, value]) => {
        const arrayIndex = key.match(/\[\d+\]$/gi);
        if (arrayIndex) {
            const index = arrayIndex[0].slice(1, -1);
            const filedName = key.substr(0, key.indexOf('['));
            if (!previous[filedName])
                previous[filedName] = [];
            previous[filedName][index] = value;
        }
        else {
            previous[key] = value;
        }
        return previous;
    }, {});
    return Object.entries(output).reduce((previous, [key, value]) => {
        if (Array.isArray(value)) {
            previous[key] = value.filter(Boolean);
            return previous;
        }
        previous[key] = value;
        return previous;
    }, {});
}

function shouldUpdateWithError({ errors, name, error, isOnBlur, isBlurType, validateDisabled, }) {
    if (validateDisabled ||
        (isOnBlur && !isBlurType) ||
        (isEmptyObject(error) && isEmptyObject(errors))) {
        return false;
    }
    if ((isEmptyObject(errors) && !isEmptyObject(error)) || (isEmptyObject(error) && errors[name])) {
        return true;
    }
    if (!errors[name]) {
        return true;
    }
    else if (errors[name].type !== error[name].type || errors[name].message !== error[name].message) {
        return true;
    }
    return false;
}

var warnMissingRef = (payload) => console.warn(`Oops missing the name for field: ${payload}`);

var modeChecker = (mode) => ({
    isOnSubmit: !mode || mode === VALIDATION_MODE.onSubmit,
    isOnBlur: mode === VALIDATION_MODE.onBlur,
    isOnChange: mode === VALIDATION_MODE.onchange,
});

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(element) {
            if (!element || !element.parentNode)
                return true;
            if (element.parentNode === window.document) {
                return false;
            }
            else if (element.parentNode === null) {
                return true;
            }
            return isDetached(element.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

var filterUndefinedErrors = (errors) => Object.entries(errors).reduce((previous, [key, value]) => {
    if (value && value.type)
        previous[key] = value;
    return previous;
}, {});

function useForm({ mode, validationSchema, defaultValues, validationFields } = {
    mode: 'onSubmit',
    defaultValues: {},
}) {
    const fieldsRef = useRef({});
    const errorsRef = useRef({});
    const submitCountRef = useRef(0);
    const touchedFieldsRef = useRef([]);
    const watchFieldsRef = useRef({});
    const isWatchAllRef = useRef(false);
    const isSubmittingRef = useRef(false);
    const isSubmittedRef = useRef(false);
    const isDirtyRef = useRef(false);
    const reRenderForm = useState({})[1];
    const validateAndStateUpdateRef = useRef();
    const renderBaseOnError = (name, errors, error) => {
        if (errors[name] && !error[name]) {
            delete errorsRef.current[name];
            reRenderForm({});
        }
        else if (error[name]) {
            reRenderForm({});
        }
    };
    const setValue = (name, value) => {
        const field = fieldsRef.current[name];
        if (!field)
            return;
        if (!touchedFieldsRef.current.includes(name)) {
            touchedFieldsRef.current.push(name);
        }
        const ref = field.ref;
        const options = field.options;
        if (isRadioInput(ref.type) && options) {
            options.forEach(({ ref: radioRef }) => {
                if (radioRef.value === value)
                    radioRef.checked = true;
            });
            return;
        }
        ref[isCheckBoxInput(ref.type) ? 'checked' : 'value'] = value;
    };
    const isValidateDisabled = () => !isSubmittedRef.current && modeChecker(mode).isOnSubmit;
    const triggerValidation = async ({ name, value, forceValidation, }) => {
        const field = fieldsRef.current[name];
        const errors = errorsRef.current;
        if (!field)
            return false;
        if (isValidateDisabled() && !forceValidation)
            return isEmptyObject(errors);
        if (value !== undefined)
            setValue(name, value);
        const error = await validateField(field, fieldsRef.current);
        errorsRef.current = Object.assign({}, filterUndefinedErrors(errorsRef.current), error);
        renderBaseOnError(name, errors, error);
        return isEmptyObject(error);
    };
    validateAndStateUpdateRef.current = validateAndStateUpdateRef.current
        ? validateAndStateUpdateRef.current
        : async ({ target: { name }, type }) => {
            if (Array.isArray(validationFields) && !validationFields.includes(name))
                return;
            const fields = fieldsRef.current;
            const errors = errorsRef.current;
            const ref = fields[name];
            if (!ref)
                return;
            const isBlurType = type === 'blur';
            const { isOnChange, isOnBlur } = modeChecker(mode);
            const validateDisabled = isValidateDisabled();
            const isWatchAll = isWatchAllRef.current;
            const shouldUpdateWatchMode = isWatchAll || watchFieldsRef.current[name];
            const shouldUpdateValidateMode = isOnChange || (isOnBlur && isBlurType);
            let shouldUpdateState = isWatchAll;
            if (!isDirtyRef.current) {
                isDirtyRef.current = true;
                shouldUpdateState = true;
            }
            if (!touchedFieldsRef.current.includes(name)) {
                touchedFieldsRef.current.push(name);
                shouldUpdateState = true;
            }
            if (validateDisabled && shouldUpdateWatchMode)
                return reRenderForm({});
            if (validationSchema) {
                const result = getFieldsValues(fields);
                const schemaValidateErrors = (await validateWithSchema(validationSchema, result)) || {};
                const error = schemaValidateErrors[name];
                const shouldUpdate = ((!error && errors[name]) || error) && shouldUpdateValidateMode;
                if (shouldUpdate || shouldUpdateWatchMode) {
                    errorsRef.current = Object.assign({}, errors, { [name]: error });
                    if (!error)
                        delete errorsRef.current[name];
                    return reRenderForm({});
                }
            }
            else {
                const error = await validateField(ref, fields);
                const shouldUpdate = shouldUpdateWithError({
                    errors,
                    error,
                    validateDisabled,
                    isOnBlur,
                    isBlurType,
                    name,
                });
                if (shouldUpdate || shouldUpdateValidateMode || shouldUpdateWatchMode) {
                    errorsRef.current = Object.assign({}, filterUndefinedErrors(errors), error);
                    renderBaseOnError(name, errorsRef.current, error);
                }
            }
            if (shouldUpdateState)
                reRenderForm({});
        };
    const removeEventListener = findRemovedFieldAndRemoveListener.bind(null, fieldsRef.current, validateAndStateUpdateRef.current);
    const setError = (name, type, message, ref) => {
        const errors = errorsRef.current;
        if (!type && errors[name]) {
            delete errors[name];
            reRenderForm({});
        }
        else if (type) {
            // can be improved with performance
            errors[name] = {
                type,
                message,
                ref,
            };
            reRenderForm({});
        }
    };
    function registerIntoFieldsRef(elementRef, data) {
        if (elementRef && !elementRef.name)
            return warnMissingRef(elementRef);
        const { name, type, value } = elementRef;
        const { required = false, validate = undefined } = data || {};
        const inputData = Object.assign({}, data, { ref: elementRef });
        const fields = fieldsRef.current;
        const isRadio = isRadioInput(type);
        const field = fields[name];
        const existRadioOptionIndex = isRadio && field && Array.isArray(field.options)
            ? field.options.findIndex(({ ref }) => value === ref.value)
            : -1;
        if ((!isRadio && field) || (isRadio && existRadioOptionIndex > -1))
            return;
        if (!type) {
            fields[name] = Object.assign({ ref: { name } }, data);
            return;
        }
        if (isRadio) {
            if (!field)
                fields[name] = { options: [], required, validate, ref: { type: 'radio', name } };
            if (validate)
                fields[name].validate = validate;
            (fields[name].options || []).push(Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) }));
        }
        else {
            fields[name] = Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) });
        }
        if (defaultValues && defaultValues[name]) {
            setValue(name, defaultValues[name]);
        }
        const fieldData = isRadio ? (fields[name].options || [])[(fields[name].options || []).length - 1] : fields[name];
        if (!fieldData)
            return;
        attachEventListeners({
            field: fieldData,
            isRadio,
            validateAndStateUpdate: validateAndStateUpdateRef.current,
        });
    }
    function watch(fieldNames, defaultValue) {
        const watchFields = watchFieldsRef.current;
        if (typeof fieldNames === 'string') {
            watchFields[fieldNames] = true;
        }
        else if (Array.isArray(fieldNames)) {
            fieldNames.forEach((name) => {
                watchFields[name] = true;
            });
        }
        else {
            isWatchAllRef.current = true;
            watchFieldsRef.current = {};
        }
        const values = getFieldsValues(fieldsRef.current, fieldNames);
        const result = values === undefined || isEmptyObject(values) ? undefined : values;
        return result === undefined ? defaultValue : result;
    }
    function register(data, rules) {
        if (!data || typeof window === 'undefined')
            return;
        if (rules && !data.name) {
            warnMissingRef(data);
            return;
        }
        if (data.name) {
            registerIntoFieldsRef(data, rules);
        }
        return (ref) => ref && registerIntoFieldsRef(ref, data);
    }
    const handleSubmit = (callback) => async (e) => {
        if (e) {
            e.preventDefault();
            e.persist();
        }
        let fieldErrors;
        let fieldValues;
        const fields = fieldsRef.current;
        const currentFieldValues = validationFields
            ? validationFields.map(name => fieldsRef.current[name])
            : Object.values(fields);
        isSubmittingRef.current = true;
        reRenderForm({});
        if (validationSchema) {
            fieldValues = getFieldsValues(fields);
            fieldErrors = await validateWithSchema(validationSchema, fieldValues);
        }
        else {
            const { errors, values } = await currentFieldValues.reduce(async (previous, field) => {
                const resolvedPrevious = await previous;
                const { ref, ref: { name }, } = field;
                if (!fields[name])
                    return Promise.resolve(resolvedPrevious);
                const fieldError = await validateField(field, fields);
                if (fieldError[name]) {
                    resolvedPrevious.errors = Object.assign({}, (resolvedPrevious.errors || {}), fieldError);
                    return Promise.resolve(resolvedPrevious);
                }
                resolvedPrevious.values[name] = getFieldValue(fields, ref);
                return Promise.resolve(resolvedPrevious);
            }, Promise.resolve({
                errors: {},
                values: {},
            }));
            fieldErrors = Object.assign({}, errors, filterUndefinedErrors(errorsRef.current));
            fieldValues = values;
        }
        isSubmittedRef.current = true;
        submitCountRef.current += 1;
        isSubmittingRef.current = false;
        if (isEmptyObject(fieldErrors)) {
            await callback(combineFieldValues(fieldValues), e);
        }
        else {
            errorsRef.current = fieldErrors;
        }
        reRenderForm({});
    };
    const unSubscribe = () => {
        fieldsRef.current &&
            Object.values(fieldsRef.current).forEach((field) => {
                const { ref, options } = field;
                isRadioInput(ref.type) && Array.isArray(options)
                    ? options.forEach((fieldRef) => removeEventListener(fieldRef, true))
                    : removeEventListener(field, true);
            });
        fieldsRef.current = {};
        watchFieldsRef.current = {};
        errorsRef.current = {};
        isWatchAllRef.current = false;
        isSubmittedRef.current = false;
        isDirtyRef.current = false;
        touchedFieldsRef.current = [];
    };
    const reset = () => {
        try {
            // @ts-ignore
            Object.values(fieldsRef.current)[0]
                .ref.closest('form')
                .reset();
        }
        catch (_a) {
            console.warn(`⚠ No HTML input found, hence <form> look up failed.`);
        }
        unSubscribe();
        reRenderForm({});
    };
    const getValues = () => getFieldsValues(fieldsRef.current);
    useEffect(() => unSubscribe, [mode]);
    return {
        register,
        handleSubmit,
        watch: watch,
        unSubscribe,
        reset,
        setError: setError,
        setValue: setValue,
        triggerValidation: triggerValidation,
        getValues,
        errors: errorsRef.current,
        formState: {
            dirty: isDirtyRef.current,
            isSubmitted: isSubmittedRef.current,
            submitCount: submitCountRef.current,
            touched: touchedFieldsRef.current,
            isSubmitting: isSubmittingRef.current,
        },
    };
}

export default useForm;
